

#include <stdlib.h>
//#include <iostream.h>
extern "C" {
  int printf(char*, ...) ;
}
extern "C" {
  #include "mpi.h"
}
#include "Seq.h"      // Not really used
#include "Procs.h"    // Not really used
#include "Threads.h"
#include "Tags.h"
#include "ProcTree.h"


void ProcTree :: synch() {
  int i ;

  for(i = 0 ; i < numChildren ; i++) {
    MPI_Irecv(NULL, 0, MPI_INT,
              child [i], BAR, MPI_COMM_WORLD, ths->event(0)) ;

    ths->waitEvent(0) ;
  }

  if(!isRoot) {
    MPI_Isend(NULL, 0, MPI_INT,
              parent, BAR, MPI_COMM_WORLD, ths->event(0)) ;

    ths->waitEvent(0) ;

    MPI_Irecv(NULL, 0, MPI_INT,
	      parent, RAB, MPI_COMM_WORLD, ths->event(0)) ;

    ths->waitEvent(0) ;
  }

  for(i = numChildren - 1 ; i >= 0 ; i--) {
    MPI_Isend(NULL, 0, MPI_INT,
	      child [i], RAB, MPI_COMM_WORLD, ths->event(0)) ;

    ths->waitEvent(0) ;
  }
}


void map2arr(int rank, int* exts,
	     char* arg1, int* arg1Strs, char* arg2, int* arg2Strs, 
	     void (*op)(void*, void*)) {
  if(rank == 0)
    (*op)(arg1, arg2) ;
  else
    for(int i = 0 ; i < *exts ; i++) {
      map2arr(rank - 1, exts + 1,
	      arg1, arg1Strs + 1, arg2, arg2Strs + 1,
	      op) ;

      arg1 += *arg1Strs ;
      arg2 += *arg2Strs ;
    }
}


void ProcTree :: red(int blockSize, char* data, void (*acc)(void*, void*)) {
  int i ;

  {
    char* temp = new char [blockSize] ;

    for(i = 0 ; i < numChildren ; i++) {
      MPI_Irecv(temp, blockSize, MPI_CHAR,
                child [i], DAT, MPI_COMM_WORLD, ths->event(0)) ;
      ths->waitEvent(0) ;

      //acc((void*) data, (void*) temp) ;
      acc(data, temp) ;
    }

    delete temp ;
  }

  if(!isRoot) {
    MPI_Isend(data, blockSize, MPI_CHAR,
	      parent, DAT, MPI_COMM_WORLD, ths->event(0)) ;
    ths->waitEvent(0) ;

    MPI_Irecv(data, blockSize, MPI_CHAR,
	      parent, DAT, MPI_COMM_WORLD, ths->event(0)) ;
    ths->waitEvent(0) ;
  }

  for(i = 0 ; i < numChildren ; i++) {
    MPI_Isend(data, blockSize, MPI_CHAR,
	      child [i], DAT, MPI_COMM_WORLD, ths->event(0)) ;
    ths->waitEvent(0) ;
  }
}


void ProcTree :: redArray(int blockSize, int rank, int* exts,
			  char* data, int* strs,
                          void (*acc)(void*, void*)) {
  int i ;

  {
    int* tempStrs = new int [rank + 1] ;
    tempStrs [0] = blockSize ;
    for(int j = 0 ; j < rank ; j++)
      tempStrs [j + 1] = exts [j] * tempStrs [j] ;
    char* temp = new char [tempStrs [rank]] ;

    MPI_Datatype* arrayType ;
    createArrayType(&arrayType, blockSize, rank, exts, tempStrs) ;

    for(i = 0 ; i < numChildren ; i++) {
      MPI_Irecv(temp, 1, *arrayType,
                child [i], DAR, MPI_COMM_WORLD, ths->event(0)) ;
      ths->waitEvent(0) ;

      map2arr(rank, exts, data, strs, temp, tempStrs, acc) ;
    }

    deleteArrayType(&arrayType, rank) ;
    delete temp ;
    delete tempStrs ;
  }

  MPI_Datatype* arrayType ;
  createArrayType(&arrayType, blockSize, rank, exts, strs) ;

  if(!isRoot) {
    MPI_Isend(data, 1, *arrayType,
	      parent, DAR, MPI_COMM_WORLD, ths->event(0)) ;
    ths->waitEvent(0) ;

    MPI_Irecv(data, 1, *arrayType,
	      parent, DAR, MPI_COMM_WORLD, ths->event(0)) ;
    ths->waitEvent(0) ;
  }

  for(i = 0 ; i < numChildren ; i++) {
    MPI_Isend(data, 1, *arrayType,
	      child [i], DAR, MPI_COMM_WORLD, ths->event(0)) ;
    ths->waitEvent(0) ;
  }

  deleteArrayType(&arrayType, rank) ;
}



void ProcTree :: redTag(int blockSize, char* data, int tagSize, char* tag,
                        int (*acc)(void*, void*)) {
  int i ;

  {
    char* tempData = new char [blockSize] ;
    char* tempTag  = new char [tagSize] ;

    for(i = 0 ; i < numChildren ; i++) {
      MPI_Irecv(tempData, blockSize, MPI_CHAR,
                child [i], DAT, MPI_COMM_WORLD, ths->event(0)) ;
      ths->waitEvent(0) ;

      MPI_Irecv(tempTag, tagSize, MPI_CHAR,
                child [i], DAT, MPI_COMM_WORLD, ths->event(0)) ;
      ths->waitEvent(0) ;

      if(acc(data, tempData))
        memcpy(tag, tempTag, tagSize) ;
    }

    delete tempTag ;
    delete tempData ;
  }

  if(!isRoot) {
    MPI_Isend(data, blockSize, MPI_CHAR,
	      parent, DAT, MPI_COMM_WORLD, ths->event(0)) ;
    ths->waitEvent(0) ;

    MPI_Isend(tag, tagSize, MPI_CHAR,
	      parent, DAT, MPI_COMM_WORLD, ths->event(0)) ;
    ths->waitEvent(0) ;

    MPI_Irecv(data, blockSize, MPI_CHAR,
	      parent, DAT, MPI_COMM_WORLD, ths->event(0)) ;
    ths->waitEvent(0) ;

    MPI_Irecv(tag, tagSize, MPI_CHAR,
	      parent, DAT, MPI_COMM_WORLD, ths->event(0)) ;
    ths->waitEvent(0) ;
  }

  for(i = 0 ; i < numChildren ; i++) {
    MPI_Isend(data, blockSize, MPI_CHAR,
	      child [i], DAT, MPI_COMM_WORLD, ths->event(0)) ;
    ths->waitEvent(0) ;

    MPI_Isend(tag, tagSize, MPI_CHAR,
	      child [i], DAT, MPI_COMM_WORLD, ths->event(0)) ;
    ths->waitEvent(0) ;
  }
}



// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    Miscellaneous global functions.                                  //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //


void createArrayType(MPI_Datatype** typeStack,
                     int blockSize, int rank, int* extents, int* strides) {

  // int extents [rank], int strides [rank]

  *typeStack = new MPI_Datatype [rank + 1] ;

  MPI_Type_contiguous(blockSize, MPI_CHAR, *typeStack + rank) ;

  for(int i = rank - 1 ; i >= 0 ; i--)
    MPI_Type_hvector(extents [i], 1, strides [i], (*typeStack) [i + 1],
                     *typeStack + i) ;

  MPI_Type_commit(*typeStack) ;  // Only supposed to use `(*typeStack) [0]'.
}


void deleteArrayType(MPI_Datatype** typeStack, int rank) {

  for(int i = 0 ; i < rank ; i++)
    MPI_Type_free(*typeStack + i) ;

  MPI_Type_free(*typeStack + rank) ;

  delete *typeStack ;
}



